# FIXME:  dbus calls time out.  Unknown yet if it's permissions or what.
use_dbus = False

import os
import stat
import locale
if use_dbus:
    import dbus
from gi.repository import Gtk
from gi.repository import GObject

import vminstall.msg
import vminstall.log
import vminstall.paths
import vminstall.caps
import vmdisks
from .tree_view import *
from .chooser import *
from vmdisks.disks import get_disk_protocols
from vmdisks.disks import protocol_file_extensions, file_extension_protocols

class DisksView(PyTreeView):
    def __init__(self, disks, renumber=True):
        self.renumber = renumber
        self._pyobj_index = 4
        self.model = Gtk.ListStore(GObject.TYPE_STRING, GObject.TYPE_STRING,
                                   GObject.TYPE_STRING, GObject.TYPE_STRING, GObject.TYPE_PYOBJECT)
        self.view = Gtk.TreeView(self.model)
        column = Gtk.TreeViewColumn(vminstall.msg.name, Gtk.CellRendererText(), text=0)
        self.view.append_column(column)
        column = Gtk.TreeViewColumn(vminstall.msg.type, Gtk.CellRendererText(), text=1)
        self.view.append_column(column)
        column = Gtk.TreeViewColumn(vminstall.msg.source, Gtk.CellRendererText(), text=2)
        self.view.append_column(column)
        self.view.set_expander_column(column)
        column = Gtk.TreeViewColumn(vminstall.msg.size_gb, Gtk.CellRendererText(), text=3)
        self.view.append_column(column)
        self.set(disks)

    def _populate_row(self, iter, disk):
        vminstall.log.debug("Adding row to disk table: '%s'" % disk)
        if self.renumber:
            disk.vdev = self.model.get_path(iter)[0]
        number = disk.get_vdev_index()
        vdev_type = disk.get_vdev_type_str()
        blocks = disk.get_virtual_size()
        if blocks is not None:
            size = '%0.1f' % (blocks/(2.0*1024*1024), )
        else:
            size = '?'
        self.model.set(iter, 0, str(number+1), 1, vdev_type, 2, disk.pdev, 3, size, 4, disk)

    def get(self):
        disks = PyTreeView.get(self)
        #disks.sort(lambda d1,d2: cmp(d1.vdev, d2.vdev))
        disks.sort(key = lambda d1: d1.vdev)
        return disks

class DisksWidget:
    def __init__(self, allow_reorder=True, remove_fn=lambda disk: True,
                 edit_fn=None,
                 add_harddisk_fn=None, add_cdrom_fn=None):
        """
        remove_fn - takes a disk that user wants to remove from VM.  Return True
                    if was removed.  Set to None instead of fn to disable remove button.
        """
        self.max_disks = 0
        self.glade = Gtk.Builder()
        self.glade.add_from_file(vminstall.paths.gladedir + "disks-window.xml")

        self.topwin = self.glade.get_object('disks-window')
        self.topwin.hide()
        self.disks_widget = self.glade.get_object('disks')
        self.disks_widget.get_parent().remove(self.disks_widget)
        self.topwin.destroy()
 
        # Right now, set renumber based on allow_reorder; maybe later expose
        # a separate setting
        self.renumber = allow_reorder
        
        s = {}

        self.support_reorder = allow_reorder
        if allow_reorder:
            s['on_button_disk_up_clicked'] = self._do_disk_up
            s['on_button_disk_down_clicked'] = self._do_disk_down
        else:
            child = self.glade.get_object('reorder-disk-alignment')
            child.get_parent().remove(child)

        self.support_remove = (remove_fn is not None)
        self.remove_fn = remove_fn
        if self.support_remove:
            s['on_button_delete_disk_clicked'] = self._do_delete_disk
        else:
            child = self.glade.get_object('button-delete-disk')
            child.get_parent().remove(child)

        self.support_edit = (edit_fn is not None)
        if edit_fn:
            s['on_button_edit_disk_clicked'] = edit_fn
        else:
            child = self.glade.get_object('button-edit-disk')
            child.get_parent().remove(child)

        self.support_add_harddisk = (add_harddisk_fn is not None)
        if add_harddisk_fn:
            s['on_button_new_harddisk_clicked'] = add_harddisk_fn
        else:
            child = self.glade.get_object('button-new-harddisk')
            child.get_parent().remove(child)

        self.support_add_cdrom = (add_cdrom_fn is not None)
        if add_cdrom_fn:
            s['on_button_new_cdrom_clicked'] = add_cdrom_fn
        else:
            child = self.glade.get_object('button-new-cdrom')
            child.get_parent().remove(child)
        self.glade.connect_signals(s)
        
        self.disks_view = DisksView([], self.renumber)
        self.glade.get_object('disks-scrolled-window').get_child().add(self.disks_view.view)
        self.disks_view.view.show_all()

    def get_widget(self):
        return self.disks_widget
        
    def set_disks(self, disks):
        vminstall.log.debug("Re-loading disks table with '%s'" % list(map(str, disks)))
        self.disks_view.set(disks)
        self.refresh()
        
    def set_max_disks(self, max_disks):
        self.max_disks = max_disks
        self.refresh()
        
    def refresh(self):
        self.glade.get_object('disks-scrolled-window').get_child().grab_focus()
        num_disks = self.disks_view.rows()
        if self.support_reorder:
            for button in ('button-disk-up', 'button-disk-down'):
                self.glade.get_object(button).set_sensitive(num_disks > 1)
        if self.support_edit:
            self.glade.get_object('button-edit-disk').set_sensitive(num_disks > 0)
        if self.support_remove:
            self.glade.get_object('button-delete-disk').set_sensitive(num_disks > 0)
        if self.support_add_cdrom:
            self.glade.get_object('button-new-cdrom').set_sensitive(num_disks < self.max_disks)
        if self.support_add_harddisk:
            self.glade.get_object('button-new-harddisk').set_sensitive(num_disks < self.max_disks)

    def get_disks(self):
        disks = self.disks_view.get()
        vminstall.log.debug("Retrieved disk list from disk table: '%s'" % list(map(str, disks)))
        return disks

    def get_selected(self):
        return self.disks_view.get_selected()

    def set_index(self, index, new_disk):
        vminstall.log.debug("Setting index %d to %s" % (index, str(new_disk)))
        self.disks_view.update(index, new_disk)
    
    def append(self, disk):
        self.disks_view.append(disk)
        self.refresh()
        
    def _do_disk_up(self, button):
        for row in self.disks_view.move_selected_up():
            disk = self.disks_view.get_row(row)
            self.disks_view.update(row, disk)
        return True

    def _do_disk_down(self, button):
        for row in self.disks_view.move_selected_down():
            disk = self.disks_view.get_row(row)
            self.disks_view.update(row, disk)
        return True

    def _do_delete_disk(self, button):
        (disk, row) = self.disks_view.get_selected()
        if disk is None:
            return True
        row = self.disks_view.delete_selected()
        assert(row is not None)
        if self.renumber:
            for i in range(row, self.disks_view.rows()):
                disk = self.disks_view.get_row(i)
                disk.vdev = i
                self.disks_view.update(i, disk)
        if self.remove_fn is not None:
            try:
                self.remove_fn(disk)
            except:
                pass
        self.refresh()
        return True


class DiskEditWidget:
    def __init__(self, virtio=False):
        self.disk = None
        self.glade = Gtk.Builder()
        self.glade.add_from_file(vminstall.paths.gladedir + "disk-edit-window.xml")
        self.topwin = self.glade.get_object('disk-edit-window')
        self.topwin.hide()
        self.disk_edit_widget = self.glade.get_object('disk-edit')
        self.disk_edit_widget.get_parent().remove(self.disk_edit_widget)
        self.topwin.destroy()

        self._protocols = vmdisks.get_disk_protocols()
        combo = self.glade.get_object('disk-protocol')
        model = Gtk.ListStore(GObject.TYPE_STRING)
        for item in self._protocols:
            item = item.rstrip(':')
            model.append([item])
        combo.set_model(model)
        combo.clear()
        cell = Gtk.CellRendererText()
        combo.pack_start(cell, True)
        combo.add_attribute(cell, 'text', 0)

        self._default_cdrom = None
        self.disk_folder = vminstall.paths.imagedir

        prompt = vminstall.msg.prompt
        bold_prompt = vminstall.msg.bold_prompt
        small = vminstall.msg.small

        def set_text(name, widget=None, text=None, modifier=None, markup=False):
            if widget is None:
                widget = self.glade.get_object(name)
            vminstall.msg.set_text(name, widget, text, modifier, markup)
        def set_radio_text(name, text=None, modifier=None, markup=False):
            widget = self.glade.get_object(name).get_children()[0]
            vminstall.msg.set_text(name, widget, text, modifier, markup)
        set_radio_text('disk-ro', text=vminstall.msg.read_only)
        set_radio_text('disk-sparse', text=vminstall.msg.create_sparse)
        set_text('disk_help')
        set_text('disk_location', text=vminstall.msg.source, modifier=prompt)
        set_text('disk_browse', text=vminstall.msg.browse)
        set_text('virtual-disk-protocol', text=vminstall.msg.storage_format, modifier=prompt)
        set_text('disk-name-prompt', text=vminstall.msg.name, modifier=prompt)
        set_text('disk_size', text=vminstall.msg.size_gb, modifier=prompt)
        set_text('cache_mode', text=vminstall.msg.cache_mode, modifier=prompt)
        set_text('virtual-disk-type', text=vminstall.msg.type, modifier=prompt)

        self.glade.connect_signals({
            'on_disk_location_changed': self.disk_location_changed,
            'on_disk_protocol_changed': self.disk_protocol_changed,
            'on_disk_cache_mode_changed': self.disk_cache_mode_changed,
            'on_disk_gb_editing_done': self.disk_gb_editing_done,
            'on_disk_sparse_toggled': self.disk_sparse_toggled,
            'on_disk_ro_toggled': self.disk_ro_toggled,
            'on_disk_browse_clicked': self.browse_existing_disk,
            'on_disk_new_clicked': self.browse_new_disk,
        })

        # set up the list for the cd-path widget
        cd_list = self.glade.get_object('disk-location')
        # Fields are raw device path, volume label, flag indicating
        # whether volume is present or not, and HAL path
        cd_model = Gtk.ListStore(str, str, bool, str)
        cd_list.set_model(cd_model)
        #cd_list.set_text_column(0)
        if not use_dbus:
            self.bus = None
            self.hal_iface = None
            self.populate_opt_media(cd_model)
        else:
            try:
                vminstall.log.debug('Getting connection to dbus.SystemBus')
                self.bus = dbus.SystemBus()
                vminstall.log.debug('Getting HAL service')
                hal_object = self.bus.get_object('org.freedesktop.Hal', '/org/freedesktop/Hal/Manager')
                self.hal_iface = dbus.Interface(hal_object, 'org.freedesktop.Hal.Manager')
                vminstall.log.debug('Enumerating CDs')
                self.populate_opt_media(cd_model)
            except Exception as e:
                vminstall.log.error("Unable to connect to HAL to list cdrom volumes: '%s'", e)
                self.bus = None
                self.hal_iface = None
                msg = Gtk.MessageDialog(parent=None, flags=0, type=Gtk.MessageType.ERROR, buttons=Gtk.ButtonsType.OK,
                                        message_format="Unable to enumerate CDROM volumes")
                msg.format_secondary_text("Cannot connect to HAL device manager")
                msg.run()
                msg.destroy()

    def default_cdrom(self):
        return self._default_cdrom or '/dev/cdrom'

    def get_widget(self):
        return self.disk_edit_widget
        
    def set(self, disk):
        self.disk = disk
        self.refresh()
        
    def get(self):
        self.disk_location_changed(update=False)
        self.disk_protocol_changed(update=False)
        self.disk_cache_mode_changed(update=False)
        self.disk_gb_editing_done()
        self.disk_sparse_toggled()
        self.disk_ro_toggled()
        self.disk.validate()
        return self.disk

    def disk_location_changed(self, widget=None, update=True):
        if not widget:
            widget = self.glade.get_object('disk-location')
        text = widget.get_active_text()
        if not text:
            return
        if self.disk.is_template() and text in (self.disk.pdev, self.disk.get_path()):
            # If it's not explicitly set, and they are setting to the default, leave it implicit.
            pass
        else:
            try:
                self.disk.pdev = text
                vminstall.log.debug("Disk location changed to '%s'" % text)
            except:
                pass
        if update:
            self.refresh(canonicalize=False)

    def update_file_extension(self, disk_image, new_proto):
        new_proto = new_proto.rsplit(":", 1)[0]
        ext = disk_image.rsplit(".", 1)
        # Is there an extension on the image name?
        if len(ext) > 1:
            path = ext[0]
            ext = ext[1]
            # Check to see if the extension is one we recognize
            if ext in file_extension_protocols:
                p = file_extension_protocols[ext]
                # Check to see if the extension matches the protocol
                if new_proto not in p:
                    if new_proto in protocol_file_extensions:
                        disk_image = path + "." + protocol_file_extensions[new_proto]
        return disk_image

    def disk_protocol_changed(self, widget=None, update=True):
        if not widget:
            widget = self.glade.get_object('disk-protocol')
        prottext = widget.get_active_text()
        if not prottext:
            return
        else:
            try:
                vminstall.log.debug("Disk protocol changed to '%s'" % prottext)
                diskwidget = self.glade.get_object('disk-location')
                disktext = diskwidget.get_active_text()
                if disktext:
                    (index, text) = self.parse_protocol()
                    if text is None:
                        text = disktext
                    else:
                        prottext += ":"
                        text = disktext.replace(text, prottext, 1)
                    text = self.update_file_extension(text, prottext)
                    self.disk.pdev = text
                    self.refresh()
            except:
                pass
        if update:
            self.refresh(canonicalize=False)

    def disk_cache_mode_changed(self, widget=None, update=True):
        if not widget:
            widget = self.glade.get_object('cache-mode-combobox')
        text = widget.get_active_text()
        if not text:
            return
        self.disk.set_cache_mode(text)

    def disk_gb_editing_done(self, widget=None):
        if not widget:
            widget = self.glade.get_object('disk-gb')
        gb = widget.get_text()
        if not gb or gb == 0:
            gb = widget.get_value()
        else:
            gb = locale.atof(gb)
        blocks = int(gb*1024*1024*2)
        self.disk.set_virtual_size(blocks)

    def disk_sparse_toggled(self, widget=None):
        if not widget:
            widget = self.glade.get_object('disk-sparse')
        self.disk.args['sparse'] = widget.get_active()

    def disk_ro_toggled(self, widget=None):
        if not widget:
            widget = self.glade.get_object('disk-ro')
        self.disk.ro = widget.get_active()

    def refresh(self, canonicalize=True):
        button = self.glade.get_object('disk-new-button')
        if (self.disk.vdevtype == vmdisks.Disk.DEVICE_DISK):
            button.show()
        else:
            button.hide()

        exists = self.disk.exists()
        self.glade.get_object('disk-name').set_text(str(self.disk.get_vdev_index()+1))
        vdev_index = list(map(lambda row: row[0], vmdisks.Disk.translations)).index(self.disk._vdevType)
        vdev_type = list(map(lambda row: row[1], vmdisks.Disk.translations))[vdev_index]
        self.glade.get_object('vdev-type').set_text(vdev_type)
        if canonicalize:
            self.glade.get_object('disk-location').get_child().set_text(self.disk.pdev)
        (index, prottext) = self.parse_protocol()
        combo = self.glade.get_object('disk-protocol')
        if prottext is None:
            prottext = "file:"
            active = 0
        combo.set_active(index)

        widget = self.glade.get_object('disk-gb')
        size_gb = widget.get_value()
        blocks = self.disk.get_virtual_size()
        if blocks is not None:
            if os.path.isdir(self.disk.get_path()):
                gb = size_gb
            else:
                gb = blocks / (2.0*1024*1024)
        else:
            if size_gb > 0:
                gb = size_gb
            else:
                gb = 4.0
        if self.disk._vdevType == vmdisks.Disk.DEVICE_CDROM:
            widget.set_sensitive(False)
        else:
            widget.set_sensitive(not exists)
        widget.set_value(gb)
        widget = self.glade.get_object('disk-ro')
        widget.set_active(self.disk.ro)
        widget.set_sensitive(self.disk._vdevType != vmdisks.Disk.DEVICE_CDROM)
        widget = self.glade.get_object('disk-sparse')
        widget.set_active('sparse' not in self.disk.args or self.disk.args['sparse'])
        widget.set_sensitive(not exists and self.disk._vdevType != vmdisks.Disk.DEVICE_CDROM)
        if vminstall.caps.is_kvm() or vminstall.caps.is_qemu():
            combo = self.glade.get_object('cache-mode-combobox')
            (index, mode) = self.disk.get_cache_mode()
            combo.set_active(index)
        else:
            self.glade.get_object("cache-mode-combobox").hide()
            self.glade.get_object("cache_mode").hide()

    def browse_existing_disk(self, ignore=None):
        self._browse_disk(mode=Gtk.FileChooserAction.OPEN)
    def browse_new_disk(self, ignore=None):
        self._browse_disk(mode=Gtk.FileChooserAction.SAVE)
    def _browse_disk(self, mode):
        file = self.glade.get_object('disk-location').get_child().get_text()
        (file, folder) = chooser(vminstall.msg.locate_disk,
                                 self.disk_edit_widget.get_toplevel(),
                                 file=file, mode=mode, folder=self.disk_folder)
        if file and folder not in file:
            check_file = folder + "/" + file
            if os.path.exists(check_file):
                file = check_file
        if file != None:
            self.disk_folder = folder
            text = self.disk.disk_image_type(file)
            if text is None:
                (index, text) = self.parse_protocol()
                if os.path.exists(file):
                    mode = os.stat(file).st_mode
                    if text and text == "phy:":
                        if stat.S_ISREG(mode):
                            # Should be whatever the default is
                            text = "file:"
                    else:
                        if stat.S_ISBLK(mode):
                            text = "phy:"
            file = text + file
            self.glade.get_object('disk-location').get_child().set_text(file)

    def parse_protocol(self):
        diskwidget = self.glade.get_object('disk-location')
        disktext = diskwidget.get_active_text()
        text = None
        index=0
        if disktext:
            while index < len(self._protocols):
                if disktext.startswith(self._protocols[index]):
                    text = self._protocols[index]
                    break
                index += 1
        if text is None:
            index = -1
        return (index, text)

    def populate_opt_media(self, model):
        if not self.bus:
            default_cdrom_local = None
            try:
                block_devs = os.listdir('/sys/block')
            except:
                return
            for bd in block_devs:
                try:
                    bdf = open('/sys/block/%s/removable' % bd)
                except:
                    continue
                if bdf.readline().strip() == "1":
                    dev = '/dev/%s' % bd
                    if 'fd' not in dev:
                        if not default_cdrom_local and dev.endswith('0'):
                            default_cdrom_local = self._default_cdrom = dev
                        if not self._default_cdrom:
                            self._default_cdrom = dev
                    model.append([dev, dev, True, dev])
            return

        vollabel = {}
        volpath = {}

        # Track device add/removes so we can detect newly inserted CD media
        self.hal_iface.connect_to_signal("DeviceAdded", self._device_added)
        self.hal_iface.connect_to_signal("DeviceRemoved", self._device_removed)

        # Find info about all current present media
        vminstall.log.debug('Querying all present media...')
        for d in self.hal_iface.FindDeviceByCapability("volume"):
            vminstall.log.debug('Getting a volume...')
            vol = self.bus.get_object("org.freedesktop.Hal", d)
            vminstall.log.debug('Got volume.')
            if vol.GetPropertyBoolean("volume.is_disc") and \
                   vol.GetPropertyBoolean("volume.disc.has_data"):
                vminstall.log.debug('a disc with data')
                devnode = vol.GetProperty("block.device")
                label = vol.GetProperty("volume.label")
                if label == None or len(label) == 0:
                    label = devnode
                vollabel[devnode] = label
                volpath[devnode] = d
            else:
                vminstall.log.debug('not a disc with data')

        vminstall.log.debug('Querying all CDROMs...')
        for d in self.hal_iface.FindDeviceByCapability("storage.cdrom"):
            dev = self.bus.get_object("org.freedesktop.Hal", d)
            devnode = dev.GetProperty("block.device")
            if devnode in vollabel:
                model.append([devnode, vollabel[devnode], True, volpath[devnode]])
            else:
                model.append([devnode, vminstall.msg.no_media, False, None])

    def _device_added(self, path):
        vol = self.bus.get_object("org.freedesktop.Hal", path)
        if vol.QueryCapability("volume"):
            if vol.GetPropertyBoolean("volume.is_disc") and \
                   vol.GetPropertyBoolean("volume.disc.has_data"):
                devnode = vol.GetProperty("block.device")
                label = vol.GetProperty("volume.label")
                if label == None or len(label) == 0:
                    label = devnode

                cdlist = self.glade.get_object("cd-path")
                model = cdlist.get_model()

                # Search for the row with matching device node and
                # fill in info about inserted media
                for row in model:
                    if row[0] == devnode:
                        row[1] = label
                        row[2] = True
                        row[3] = path

    def _device_removed(self, path):
        vol = self.bus.get_object("org.freedesktop.Hal", path)
        cdlist = self.glade.get_object("cd-path")
        model = cdlist.get_model()

        active = cdlist.get_active()
        idx = 0
        # Search for the row containing matching HAL volume path
        # and update (clear) it, de-activating it if its currently
        # selected
        for row in model:
            if row[3] == path:
                row[1] = vminstall.msg.no_media
                row[2] = False
                row[3] = None
                if idx == active:
                    cdlist.set_active(-1)
            idx = idx + 1

